/*
 * Copyright (C) 2000-2011 OSS Nokalva.  All rights reserved.
 *
 * THIS FILE IS PROPRIETARY MATERIAL OF OSS NOKALVA AND MAY BE USED ONLY
 * BY DIRECT LICENSEES OF OSS NOKALVA. THIS FILE MAY NOT BE DISTRIBUTED.
 */

/*
 * This program is designed for checking files containing BER/DER definite-
 * length encoded UNIQUE PDUs.  The uniqueness of all the PDUs in the file is
 * an important condition for the code to succeed. Another important condition
 * is the first tag-length pairs being valid, otherwise the result is
 * unpredictable.  The code reports which concatenated encoded PDUs have
 * errors.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "ctltbl.h"

static int getPDUFromFile(OssGlobal *world, FILE *fp, OssBuf *encodedPDU);



int main(int argc, char **argv)
{

    int		retcode, error = 0;
    OssBuf	encodedPDU;
    void	*decodedPDU = NULL;
    int		pdu, pdu_num = 0;
    OssGlobal	w, *world = &w;
    FILE	*fp;

	if (argc != 2) {
        printf("Usage: %s <encoded-PDU-data-file>\n",argv[0]);
        return 1;
    }

    if (retcode = ossinit(world, ctltbl)) {
	printf("Ossinit() returned %d\n", retcode);
	return retcode;
    }

    fp = fopen(argv[1], "rb");
    if (!fp) {
	ossPrint(world, "Can't open an encoded PDU file.\n");
	return 1;
    }

    for (pdu = 1; !retcode; pdu++) {
	retcode = getPDUFromFile(world, fp, &encodedPDU);
	if (retcode == 1) {
	    ossPrint(world, "Can't get a PDU #%d from file.\n", pdu);
	    continue;
	}
	else if (retcode == 2) {
	    if (!error)
	        ossPrint(world, "All PSUs are valid BER/DER encodings.\n");
	    else
	        ossPrint(world, "Some PDUs are invalid.\n");
	    continue;
	}
	if (ossDecode(world, &pdu_num, &encodedPDU, &decodedPDU)) {
	    ossPrint(world, "Concatenated PDU #%d is ivalid.\n", pdu);
	    ossPrint(world, "%s", ossGetErrMsg(world));
	    error = 1;
	}
	free(encodedPDU.value);
	decodedPDU = NULL;
	pdu_num = 0;
    }

    fclose(fp);
    ossterm(world);
    return retcode == 1? retcode: 0;
}


/*
 * Function:            reads an encoded PDU from a file and returns it in
 *                      the OssBuf.  For INDEFINITE-length encodings an error
 *                      code is returned.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's global structure
 *   "fp"           ... file pointer
 *   "encodedPDU"   ... encoded PDU OssBuf
 *
 * Return value:        zero if successful and error code on error
 */
static int getPDUFromFile(OssGlobal *world, FILE *fp, OssBuf *encodedPDU)
{

    size_t	length;	/* size of encoded tag & length to skip */
    int		byte;
    unsigned char buffer[8], *buf = buffer;


	/*
	 * Read first octet of encoded tag (issue an error if failed).
	 */
    if ((byte = fgetc(fp)) == EOF) {
	if (feof(fp))
	    return 2;
	else {
	    ossPrint(world, "Error reading file.\n");
	    return 1;
	}
    }
    *buf++ = byte;
    length = 1;

	/*
	 * If five lower bits are all 1's, the encoding of tag is a long form.
	 * Skipping all subsequent octets with highest bits set.
	 */
    if ((byte & 0x1F) == 0x1F) {
	do {
	    if ((byte = fgetc(fp)) == EOF) {
		if (feof(fp))
		    return 1;
		else {
		    ossPrint(world, "Error reading the file.\n");
		    return 1;
		}
	    }
	    *buf++ = byte;
	    length++;		/* count this octet */
	} while (byte & 0x80);	/* this one was the last octet of the tag */
    }
    
    if ((byte = fgetc(fp)) == EOF) {
	if (feof(fp))
	    return 1;
	else {
	    ossPrint(world, "Error reading file.\n");
	    return 1;
	}
    }
    length++;		/* skip this octet */
    *buf++ = byte;

    if (byte == 0x80) {	/* indicates INDEFINITE length */
	encodedPDU->length = -1;	/* means INDEFINITE */
	return 1;
    }
    else {
	if (byte & 0x80) { /* higher bit set indicates long form */
	    int len_of_len = byte & 0x7F;
	    
	    length += len_of_len;	/* number bytes used for subsequent octets of "length" */
	    encodedPDU->length = 0;
	    while (len_of_len--) {
		if ((byte = fgetc(fp)) == EOF) {
		    if (feof(fp))
			return 1;
		    else {
			ossPrint(world, "Error reading file.\n");
			return 1;
		    }
		}
		*buf++ = byte;
		length++;
		encodedPDU->length = ((unsigned long)encodedPDU->length << 8) + byte;
	    }
	}
	else
	    encodedPDU->length = byte;
    }

    encodedPDU->value = malloc(encodedPDU->length + length);
    if (!encodedPDU->value)
	return 1;
    memcpy(encodedPDU->value, buffer, length);
    if (fread(encodedPDU->value + length, 1, encodedPDU->length, fp) != (size_t)encodedPDU->length) {
	if (feof(fp))
	    return 1;
	else {
	    ossPrint(world, "Error reading file.\n");
	    return 1;
	}
    }
    encodedPDU->length += length;
    return 0;
}
